import json
from enum import Enum
from typing import Any


def default_serializer(obj: Any) -> str:
    """Smart serializer that handles various Python types appropriately.
    
    - Complex data structures (lists, dicts) use JSON
    - Custom objects try to convert to dict or string representation
    - Handles dataclasses, enums, and other common types
    """
    # Handle None explicitly
    if obj is None:
        return "null"

    # For strings, just return them within JSON (keeps quotes and escaping)
    # For other primitives, use standard JSON serialization
    if isinstance(obj, (str, int, float, bool, type(None))):
        return json.dumps(obj)

    # Complex data structures use pretty JSON
    try:
        # Custom encoder function for handling non-standard JSON types
        def json_encoder(o):
            # Handle dataclasses
            if hasattr(o, "__dataclass_fields__"):
                return {f: getattr(o, f) for f in o.__dataclass_fields__}

            # Handle Enums
            if isinstance(o, Enum):
                return o.value

            # Handle objects with __dict__
            if hasattr(o, "__dict__"):
                return o.__dict__

            # Last resort: convert to string
            return str(o)

        return json.dumps(obj, default=json_encoder, ensure_ascii=False)
    except:
        # Fallback to basic string representation if JSON fails
        return f'"{str(obj)}"'  # Quote the string to make valid JSON


def default_exception_serializer(exc: Exception) -> str:
    """Default serializer for exceptions.
    
    Returns a JSON object with the exception type and message.
    """
    return json.dumps({
        "error": {
            # "type": type(exc).__name__,
            "message": str(exc),
            # Include traceback for debugging but without full paths
            # "traceback": traceback.format_exc().splitlines()
        }
    })



import inspect
import warnings
from dataclasses import is_dataclass, fields, MISSING
from enum import Enum
from types import UnionType
from typing import (
    Any,
    Dict,
    get_origin,
    Annotated,
    get_args,
    Literal,
    Union,
    List,
    Tuple,
    Dict as TypingDict,
    NamedTuple,
    Type,
    cast,
)

NoneType = type(None)

type_map = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    NoneType: "null",
}


def is_named_tuple_type(t: Any) -> bool:
    return (
        isinstance(t, type)
        and issubclass(t, tuple)
        and hasattr(t, "__annotations__")
        and hasattr(t, "_fields")
    )


def is_typed_dict(obj: Any) -> bool:
    return (
        isinstance(obj, type)
        and issubclass(obj, dict)
        and hasattr(obj, "__annotations__")
        and isinstance(getattr(obj, "__annotations__", None), dict)
    )


def is_autogenerated_namedtuple_doc(namedtuple_cls: Type[NamedTuple]) -> bool:
    expected_doc: str = (
        f"{namedtuple_cls.__name__}({', '.join(namedtuple_cls._fields)})"
    )
    return namedtuple_cls.__doc__ == expected_doc


def python_type_to_json_schema(py_type: Any, strict: bool = False) -> Dict[str, Any]:
    origin = get_origin(py_type)

    # Initialize base schema with ordered keys
    schema: Dict[str, Any] = {}

    # Handle Annotated by extracting description first
    description = None
    if origin is Annotated:
        args = get_args(py_type)
        underlying_type = args[0]
        metadata = args[1:]
        if metadata and isinstance(metadata[0], str):
            description = metadata[0]
        py_type = underlying_type
        origin = get_origin(underlying_type)

    # Handle Literal
    if origin is Literal:
        values = get_args(py_type)
        first_type = type(values[0])
        if all(isinstance(v, first_type) for v in values):
            schema["type"] = type_map.get(first_type, None)
        if description:
            schema["description"] = description
        schema["enum"] = list(values)
        return schema

    # Enums
    if inspect.isclass(py_type) and issubclass(py_type, Enum):
        schema["type"] = "string"
        if description:
            schema["description"] = description
        schema["enum"] = [member.value for member in py_type]
        return schema

    # Union (including Optional)
    if origin is Union or isinstance(py_type, UnionType):
        args = get_args(py_type)
        # Optional[T]
        if len(args) == 2 and NoneType in args:
            non_none_type = args[0] if args[1] is NoneType else args[1]
            sub_schema = python_type_to_json_schema(non_none_type, strict=strict)
            # Copy all properties from sub_schema
            for key, value in sub_schema.items():
                if key != "type" and key != "oneOf":
                    schema[key] = value
            
            # Always use type array approach when possible
            non_none_schema_type = sub_schema.get("type")
            if non_none_schema_type:
                if isinstance(non_none_schema_type, list):
                    if "null" not in non_none_schema_type:
                        schema["type"] = non_none_schema_type + ["null"]
                else:
                    schema["type"] = [non_none_schema_type, "null"]
            elif "oneOf" in sub_schema:
                # For complex unions, still use oneOf
                schema["oneOf"] = sub_schema["oneOf"] + [{"type": "null"}]
            elif "enum" in sub_schema:
                # For enums, preserve the enum values and add null to type
                schema["enum"] = sub_schema["enum"]
                schema["type"] = ["string", "null"]
            else:
                # If no type and no oneOf, default to string|null
                schema["type"] = ["string", "null"]
            
            if description:
                schema["description"] = description
            return schema
        else:
            if description:
                schema["description"] = description
            schema["oneOf"] = [
                python_type_to_json_schema(a, strict=strict) for a in args
            ]
            return schema

    # Lists (List[T])
    if origin in (list, List):
        schema["type"] = "array"
        if description:
            schema["description"] = description
        (item_type,) = get_args(py_type)
        schema["items"] = python_type_to_json_schema(item_type, strict=strict)
        return schema

    # Tuples
    if origin in (tuple, Tuple):
        schema["type"] = "array"
        if description:
            schema["description"] = description

        args_ = get_args(py_type)
        if len(args_) == 2 and args_[1] is Ellipsis:
            # variable-length tuple
            item_schema = python_type_to_json_schema(args_[0], strict=strict)
            schema["items"] = item_schema
        else:
            # Fixed-length tuple using prefixItems
            warnings.warn(
                "Fixed-length tuples have limited schema support and validation. "
                "Consider using NamedTuple instead for better type safety and schema generation."
            )
            items = [python_type_to_json_schema(a, strict=strict) for a in args_]
            length = len(items)
            schema["prefixItems"] = items
            schema["items"] = False  # Disallow additional items
            schema["minItems"] = length
            schema["maxItems"] = length
        return schema

    # Dicts (Dict[str, T])
    if origin in (dict, TypingDict):
        schema["type"] = "object"
        if description:
            schema["description"] = description
        key_type, value_type = get_args(py_type)
        schema["additionalProperties"] = python_type_to_json_schema(
            value_type, strict=strict
        )
        if strict:
            schema["additionalProperties"] = False
        return schema

    # TypedDict
    if is_typed_dict(py_type):
        schema["type"] = "object"
        if description:
            schema["description"] = description
        elif py_type.__doc__:
            schema["description"] = py_type.__doc__.strip()

        props = {}
        td_annotations = py_type.__annotations__
        for k, v in td_annotations.items():
            props[k] = python_type_to_json_schema(v, strict=strict)
        schema["properties"] = props

        required_fields = list(props.keys())
        if hasattr(py_type, "__optional_keys__"):
            for opt_key in py_type.__optional_keys__:
                if opt_key in required_fields:
                    required_fields.remove(opt_key)
        if strict:
            schema["additionalProperties"] = False
        if required_fields:
            schema["required"] = required_fields

        return schema

    # NamedTuple
    if is_named_tuple_type(py_type):
        schema["type"] = "object"
        if description:
            schema["description"] = description
        elif py_type.__doc__ and not is_autogenerated_namedtuple_doc(
            cast(type[NamedTuple], py_type)
        ):
            schema["description"] = py_type.__doc__.strip()

        nt_fields = py_type.__annotations__
        props = {
            fname: python_type_to_json_schema(ftype, strict=strict)
            for fname, ftype in nt_fields.items()
        }
        if strict:
            schema["additionalProperties"] = False
        schema["properties"] = props
        schema["required"] = list(props.keys())
        return schema

    # Dataclass
    if is_dataclass(py_type):
        schema["type"] = "object"
        if description:
            schema["description"] = description
        elif py_type.__doc__ and not py_type.__doc__.startswith(py_type.__name__):
            schema["description"] = py_type.__doc__.strip()

        props = {}
        required = []
        for f in fields(py_type):
            f_schema = python_type_to_json_schema(f.type)
            props[f.name] = f_schema
            if f.default is MISSING and f.default_factory is MISSING:
                required.append(f.name)
        schema["properties"] = props
        if required:
            schema["required"] = required
        if strict:
            schema["additionalProperties"] = False
        return schema

    # Any
    if py_type is Any:
        if description:
            schema["description"] = description
        return schema

    # Primitive fallback
    schema["type"] = type_map.get(py_type, "string")
    if description:
        schema["description"] = description
    return schema


def json_to_python(value: Any, py_type: Any) -> Any:
    origin = get_origin(py_type)
    if origin is Annotated:
        py_type = get_args(py_type)[0]
        origin = get_origin(py_type)

    if origin is Literal:
        return value

    if isinstance(py_type, type) and issubclass(py_type, Enum):
        return py_type(value)

    if origin is Union or isinstance(py_type, UnionType):
        args = get_args(py_type)
        if len(args) == 2 and NoneType in args:
            if value is None:
                return None
            non_none_type = args[0] if args[1] is NoneType else args[1]
            return json_to_python(value, non_none_type)
        else:
            for candidate in args:
                try:
                    return json_to_python(value, candidate)
                except Exception:
                    pass
            raise ValueError("No union variant matched the value")

    if origin in (list, List):
        (item_type,) = get_args(py_type)
        return [json_to_python(v, item_type) for v in value]

    if origin in (tuple, Tuple):
        args_ = get_args(py_type)
        if len(args_) == 2 and args_[1] is Ellipsis:
            item_type = args_[0]
            return tuple(json_to_python(v, item_type) for v in value)
        else:
            if len(value) != len(args_):
                raise ValueError("Tuple length mismatch")
            return tuple(json_to_python(v, t) for v, t in zip(value, args_))

    if origin in (dict, TypingDict):
        key_type, val_type = get_args(py_type)
        return {k: json_to_python(v, val_type) for k, v in value.items()}

    if is_typed_dict(py_type):
        td_annotations = py_type.__annotations__
        result = {}
        for k, val_type in td_annotations.items():
            if k in value:
                result[k] = json_to_python(value[k], val_type)
        return result

    if is_named_tuple_type(py_type):
        nt_fields = py_type.__annotations__
        kwargs = {
            fname: json_to_python(value[fname], ftype)
            for fname, ftype in nt_fields.items()
        }
        return py_type(**kwargs)

    if is_dataclass(py_type):
        cls_fields = fields(py_type)
        kwargs = {}
        for f in cls_fields:
            if f.name in value:
                kwargs[f.name] = json_to_python(value[f.name], f.type)
            elif f.default is not MISSING:
                kwargs[f.name] = f.default
            elif f.default_factory is not MISSING:
                kwargs[f.name] = f.default_factory()
            else:
                raise ValueError(f"Missing required field {f.name}")
        return py_type(**kwargs)

    return value



from toolkitr._registry import ToolRegistry
from toolkitr._serializer import default_serializer, default_exception_serializer

__all__ = ["ToolRegistry", "default_serializer", "default_exception_serializer"]



import inspect
import json
import asyncio
from typing import (
    Any,
    Callable,
    get_type_hints,
    Optional,
    Literal,
    TypedDict,
    Iterator, Coroutine,
    Union, get_origin, get_args,
)
from toolkitr._schema import NoneType
from dataclasses import dataclass, field

from toolkitr._schema import python_type_to_json_schema, json_to_python
from toolkitr._serializer import default_exception_serializer


class ToolFunctionDefinition(TypedDict, total=False):
    """A definition of a function tool."""

    name: str
    description: str
    parameters: dict[str, Any]
    strict: bool
    title: str  # Human-friendly name


class ToolDefinition(TypedDict):
    """A definition of a registered tool."""

    type: Literal["function"]
    function: ToolFunctionDefinition


@dataclass(frozen=True)
class ToolInfo:
    """Information about a registered tool."""

    name: str
    description: str
    parameters: dict[str, Any]  # JSON Schema for parameters
    function: Callable
    is_async: bool = False
    strict: bool = False
    title: Optional[str] = None  # Human-friendly name
    response_serializer: Optional[Callable[[Any], str]] = None
    exception_serializer: Optional[Callable[[Exception], str]] = None

    def __post_init__(self):
        if self.title is None:
            # Set title to function name if not provided, as it is frozen we need to use object.__setattr__
            object.__setattr__(self, "title", self.name)


    @property
    def definition(self) -> ToolDefinition:
        function_def = {
            "name": self.name,
            "description": self.description,
            "parameters": self.parameters,
            "strict": self.strict,
        }
        
        # Add title to the definition if it exists
        if self.title:
            function_def["title"] = self.title
            
        return {
            "type": "function",
            "function": function_def,
        }


class ToolCallFunctionDict(TypedDict):
    name: str
    arguments: str  # In JSON format


class ToolCallDict(TypedDict):
    type: Literal["function"]
    id: str
    function: ToolCallFunctionDict


class ToolCallMessageDict(TypedDict):
    """Message returned from a tool call execution."""

    role: Literal["tool"]
    tool_call_id: str
    name: str
    content: str


@dataclass
class ToolCallResult:
    """Result of a tool call execution."""
    
    # Information about the tool that was called
    tool: ToolInfo
    
    # The formatted response message for LLM
    message: ToolCallMessageDict
    
    # The result from the tool function (None if it failed)
    result: Optional[Any] = None
    
    # The exception if the call failed (None if successful)
    error: Optional[Exception] = None
    
    @property
    def success(self) -> bool:
        """Returns True if the tool call was successful."""
        return self.error is None


class ToolRegistry:
    def __init__(
        self, 
        strict: bool = False,
        response_serializer: Callable[[Any], str] = None,
        exception_serializer: Callable[[Exception], str] = None,
        catch_exceptions: bool = True
    ):
        self._registry: dict[str, ToolInfo] = {}
        self._default_strict = strict
        self._response_serializer = response_serializer or json.dumps
        self._exception_serializer = exception_serializer or default_exception_serializer
        self._catch_exceptions = catch_exceptions
        
    @property
    def response_serializer(self) -> Callable[[Any], str]:
        return self._response_serializer
        
    @response_serializer.setter
    def response_serializer(self, serializer: Callable[[Any], str]):
        self._response_serializer = serializer
        
    @property
    def exception_serializer(self) -> Callable[[Exception], str]:
        return self._exception_serializer
        
    @exception_serializer.setter
    def exception_serializer(self, serializer: Callable[[Exception], str]):
        self._exception_serializer = serializer
        
    @property
    def catch_exceptions(self) -> bool:
        return self._catch_exceptions
        
    @catch_exceptions.setter
    def catch_exceptions(self, value: bool):
        self._catch_exceptions = value

    def __getitem__(self, tool_name: str):
        return self._registry[tool_name]

    def __iter__(self) -> Iterator[ToolInfo]:
        return iter(self._registry.values())

    def __len__(self) -> int:
        return len(self._registry)

    def __contains__(self, item) -> bool:
        return item in self._registry

    def import_registry(
        self, 
        registry: 'ToolRegistry', 
        namespace: Optional[str] = None,
        overwrite: bool = False
    ):
        """Import tools from another registry.
        
        Args:
            registry: The registry to import tools from
            namespace: Optional prefix to add to imported tool names (e.g., "math")
            overwrite: Whether to overwrite existing tools with the same name
        
        Example:
            # Create registries
            math_registry = ToolRegistry()
            math_registry.register_tool(add)
            math_registry.register_tool(subtract)
            
            main_registry = ToolRegistry()
            main_registry.import_registry(math_registry, namespace="math")
            
            # Call imported tools
            result = main_registry.call("math.add", a=1, b=2)  # Returns 3
        """
        for tool_info in registry:
            if namespace:
                name = f"{namespace}.{tool_info.name}"
            else:
                name = tool_info.name
                
            if name in self and not overwrite:
                raise ValueError(f"Tool with name '{name}' already exists. Use overwrite=True to replace it.")
            
            # Create a new ToolInfo with the updated name
            imported_tool = ToolInfo(
                name=name,
                description=tool_info.description,
                parameters=tool_info.parameters,
                function=tool_info.function,
                is_async=tool_info.is_async,
                strict=tool_info.strict,
                title=tool_info.title,
                response_serializer=tool_info.response_serializer,
                exception_serializer=tool_info.exception_serializer,
            )
            
            # Add to registry
            self._registry[name] = imported_tool

    def register_tool(
        self,
        func: Callable,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        title: Optional[str] = None,
        strict: Optional[bool] = None,
        response_serializer: Optional[Callable[[Any], str]] = None,
        exception_serializer: Optional[Callable[[Exception], str]] = None,
    ):
        strict = self._default_strict if strict is None else strict
        sig = inspect.signature(func)
        type_hints = get_type_hints(func, include_extras=True)

        tool_name = name if name else func.__name__
        tool_description = description if description else (func.__doc__ or "").strip()

        properties = {}
        required_fields = []
        for param_name, param in sig.parameters.items():
            if param_name == "self":
                continue
            py_type = type_hints.get(param_name, Any)
            schema_prop = python_type_to_json_schema(py_type, strict=strict)
            properties[param_name] = schema_prop
            
            # In strict mode, add all parameters to required fields
            # In non-strict mode, only add if not optional and has no default
            if strict:
                required_fields.append(param_name)
            else:
                # Check if the parameter is optional
                is_optional = False
                origin = get_origin(py_type)
                if origin is Union:
                    args = get_args(py_type)
                    if len(args) == 2 and NoneType in args:
                        is_optional = True
                
                # Only add to required if it has no default value and is not Optional
                if param.default is inspect.Parameter.empty and not is_optional:
                    required_fields.append(param_name)

        parameters_schema = {"type": "object", "properties": properties}
        if strict:
            parameters_schema["additionalProperties"] = False
        if required_fields:
            parameters_schema["required"] = required_fields

        is_async = inspect.iscoroutinefunction(func)

        tool_info = ToolInfo(
            name=tool_name,
            description=tool_description,
            parameters=parameters_schema,
            function=func,
            is_async=is_async,
            strict=strict,
            title=title,
            response_serializer=response_serializer,
            exception_serializer=exception_serializer,
        )

        self._registry[tool_name] = tool_info

    @staticmethod
    def _build_arguments(func: Callable, arguments: dict[str, Any]) -> dict[str, Any]:
        sig = inspect.signature(func)
        type_hints = get_type_hints(func, include_extras=True)
        py_kwargs = {}
        for param_name, param in sig.parameters.items():
            if param_name == "self" and inspect.ismethod(func):
                continue
            py_type = type_hints.get(param_name, Any)
            if param_name in arguments:
                py_kwargs[param_name] = json_to_python(arguments[param_name], py_type)
        return py_kwargs

    def _handle_exception(self, exc: Exception, tool_info: ToolInfo) -> str:
        """Handle an exception from a tool call."""
        if not self._catch_exceptions:
            raise exc
        serializer = tool_info.exception_serializer or self._exception_serializer
        return serializer(exc)
        
    def _create_tool_response(
        self, 
        tool_call_id: str, 
        name: str, 
        content: str
    ) -> ToolCallMessageDict:
        """Create a tool call response."""
        return {
            "role": "tool",
            "tool_call_id": tool_call_id,
            "name": name,
            "content": content,
        }
        
    def call(self, name: str, **arguments: any):
        tool_info = self[name]
        func = tool_info.function
        py_kwargs = self._build_arguments(func, arguments)
        return func(**py_kwargs)

    async def acall(self, name: str, **arguments: any):
        tool_info = self[name]
        func = tool_info.function
        py_kwargs = self._build_arguments(func, arguments)
        return await func(**py_kwargs)

    def tool_call(self, tool_call: ToolCallDict) -> ToolCallResult:
        """Execute a synchronous tool call.
        
        Args:
            tool_call: The tool call dictionary in OpenAI format
            
        Returns:
            A ToolCallResult object containing the result or error, tool info, and formatted message
        """
        function = tool_call["function"]
        function_name = function["name"]
        tool = self[function_name]
        tool_call_id = tool_call["id"]
        
        result = None
        error = None
        
        try:
            arguments = json.loads(function["arguments"])
            result = self.call(function_name, **arguments)
            
            # Use tool-specific serializer if available, otherwise use registry default
            serializer = tool.response_serializer or self._response_serializer
            content = serializer(result)
        except Exception as exc:
            error = exc
            content = self._handle_exception(exc, tool)
            
        message = self._create_tool_response(tool_call_id, function_name, content)
        
        return ToolCallResult(
            result=result,
            error=error,
            tool=tool,
            message=message
        )

    async def atool_call(self, tool_call: ToolCallDict) -> ToolCallResult:
        """Execute an asynchronous tool call.
        
        Args:
            tool_call: The tool call dictionary in OpenAI format
            
        Returns:
            A ToolCallResult object containing the result or error, tool info, and formatted message
        """
        function = tool_call["function"]
        function_name = function["name"]
        tool = self[function_name]
        tool_call_id = tool_call["id"]
        
        result = None
        error = None
        
        try:
            arguments = json.loads(function["arguments"])
            result = await self.acall(function_name, **arguments)
            
            # Use tool-specific serializer if available, otherwise use registry default
            serializer = tool.response_serializer or self._response_serializer
            content = serializer(result)
        except Exception as exc:
            error = exc
            content = self._handle_exception(exc, tool)
            
        message = self._create_tool_response(tool_call_id, function_name, content)
        
        return ToolCallResult(
            result=result,
            error=error,
            tool=tool,
            message=message
        )
        
    async def smart_call(self, name: str, **arguments: any):
        """Call a tool regardless of whether it's sync or async.
        
        This method automatically detects if the tool is synchronous or asynchronous
        and calls it appropriately. Synchronous tools are executed in a thread pool
        to prevent blocking the event loop.
        
        Args:
            name: The name of the tool to call
            **arguments: The arguments to pass to the tool
            
        Returns:
            The result of the tool execution
        """
        tool_info = self[name]
        if tool_info.is_async:
            return await self.acall(name, **arguments)
        else:
            # Run synchronous function in thread pool to prevent blocking
            return await asyncio.to_thread(self.call, name, **arguments)
    
    async def smart_tool_call(self, tool_call: ToolCallDict) -> ToolCallResult:
        """Handle tool calls for both sync and async tools automatically.
        
        This method automatically detects if the tool is synchronous or asynchronous
        and processes the tool call appropriately. Synchronous tools are executed in
        a thread pool to prevent blocking the event loop.
        
        Args:
            tool_call: The tool call dictionary in OpenAI format
            
        Returns:
            A ToolCallResult object containing the result or error, tool info, and formatted message
        """
        function_name = tool_call["function"]["name"]
        tool = self[function_name]
        tool_call_id = tool_call["id"]
        
        result = None
        error = None
        
        try:
            arguments = json.loads(tool_call["function"]["arguments"])
            
            if tool.is_async:
                result = await self.acall(function_name, **arguments)
            else:
                # Run synchronous function in thread pool
                result = await asyncio.to_thread(self.call, function_name, **arguments)
            
            # Use tool-specific serializer if available, otherwise use registry default
            serializer = tool.response_serializer or self._response_serializer
            content = serializer(result)
        except Exception as exc:
            error = exc
            content = self._handle_exception(exc, tool)
        
        message = self._create_tool_response(tool_call_id, function_name, content)
        
        return ToolCallResult(
            result=result,
            error=error,
            tool=tool,
            message=message
        )

    def definitions(self) -> list[dict[str, Any]]:
        return [tool_info.definition for tool_info in self._registry.values()]

    def tool(
        self,
        *,
        name: str = None,
        description: str = None,
        title: str = None,
        strict: Optional[bool] = None,
        response_serializer: Optional[Callable[[Any], str]] = None,
        exception_serializer: Optional[Callable[[Exception], str]] = None,
    ):
        def decorator(func: Callable):
            self.register_tool(
                func, 
                name=name, 
                description=description,
                title=title,
                strict=strict,
                response_serializer=response_serializer,
                exception_serializer=exception_serializer
            )
            return func

        return decorator
